'''
Created on 2023-07-30

@author: wf
'''
import unittest
from tests.basetest import Basetest
from nicescad.solidservice import SolidConverter, FastAPIServer
from fastapi.testclient import TestClient
import nicescad as nicescad

class TestSolidConverter(Basetest):
    """
    Unit tests for the OpenSCAD code converter (SolidConverter class in the nicescad.solidservice module).

    The test_conversion function checks whether the SolidConverter returns the
    expected OpenSCAD code for a given Python code input.

    These tests are for https://github.com/WolfgangFahl/nicescad/issues/28.

    Attributes:
    Author: OpenAI GPT-4
    Date: July 30, 2023
    """

    def setUp(self):
        super().setUp(debug=False, profile=True)

    def test_conversion(self):
        """
        Test whether the SolidConverter returns the expected OpenSCAD code for a 
        given Python code input.
        """
        # The Python code to be converted
        python_code = "difference()(cube(10),sphere(15))"

        # The expected OpenSCAD code
        expected_openscad_code = """difference(){cube(size=10);sphere(r=15);}"""
        # Initialize a SolidConverter object with the Python code
        converter = SolidConverter(python_code)

        # Assert that the conversion result matches the expected OpenSCAD code
        scad=converter.convert_to_openscad()
        debug=self.debug
        if debug:
            print(scad)
        scad = "".join(scad.split())
        if debug:
            print(scad)
        self.assertEqual(scad, expected_openscad_code)
        
    def check_server(self):
        """
        Check whether the FastAPI server is running and returns the expected version of the nicescad package.
    
        Returns:
        bool -- True if the server is running and the version is correct, False otherwise.
        """
        try:
            response = self.client.get("/version/")
            if response.status_code != 200:
                return False
            data = response.json()
            if 'version' not in data or data['version'] != nicescad.__version__:
                return False
            return True
        except:
            return False
    
    def test_endpoint(self):
        """
        Test whether the FastAPI endpoint returns the expected OpenSCAD code for a 
        given Python code input.
        """
        self.server = FastAPIServer()
        self.client = TestClient(self.server.app)
        if not self.check_server():
            print("Server not available")
            return
        
        python_code = "difference()(cube(10),sphere(15))"
        expected_openscad_code = """difference(){cube(size=10);sphere(r=15);}"""

        response = self.client.post("/convert/", json={"python_code": python_code})

        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertIn('openscad_code', data)
        received_openscad_code = "".join(data['openscad_code'].split())

        self.assertEqual(received_openscad_code, expected_openscad_code)


if __name__ == '__main__':
    unittest.main()



"""
Created on 2023-06-19

@author: wf
"""
import os
from pathlib import Path

from ngwidgets.file_selector import FileSelector
from ngwidgets.input_webserver import InputWebserver
from ngwidgets.local_filepicker import LocalFilePicker
from ngwidgets.webserver import WebserverConfig
from nicegui import app, ui
from nicegui.events import ColorPickEventArguments

from nicescad.axes_helper import AxesHelper
from nicescad.openscad import OpenScad
from nicescad.version import Version


class WebServer(InputWebserver):
    """WebServer class that manages the server and handles OpenScad operations.

    Attributes:
        oscad (OpenScad): An OpenScad object that aids in performing OpenScad operations.
    """

    @classmethod
    def get_config(cls) -> WebserverConfig:
        copy_right = "(c)2023 Wolfgang Fahl"
        config = WebserverConfig(
            copy_right=copy_right, version=Version(), default_port=9858
        )
        return config

    def __init__(self):
        """Constructs all the necessary attributes for the WebServer object."""
        InputWebserver.__init__(self, config=WebServer.get_config())
        self.oscad = OpenScad(
            scad_prepend="""//https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fa,_$fs_and_$fn
// default number of facets for arc generation
$fn=30;
"""
        )
        self.code = """// nicescad example
module example() {
  translate([0,0,15]) {
     cube(30,center=true);
     sphere(20);
  }
}
example();"""
        self.input = "example.scad"
        self.stl_name = "result.stl"
        self.stl_color = "#57B6A9"
        self.stl_object = None
        app.add_static_files("/stl", self.oscad.tmp_dir)
        self.do_trace = True
        self.html_view = None
        self.axes_view = None

        @ui.page("/")
        async def home():
            await self.home()

        @ui.page("/settings")
        async def settings():
            await self.settings()

    @classmethod
    def examples_path(cls) -> str:
        # the root directory (default: examples)
        path = os.path.join(os.path.dirname(__file__), "../nicescad_examples")
        path = os.path.abspath(path)
        return path

    async def render(self, _click_args=None):
        """Renders the OpenScad string and updates the 3D scene with the result.

        Args:
            click_args (object): The click event arguments.
        """
        try:
            self.progress_view.visible = True
            ui.notify("rendering ...")
            with self.scene:
                self.stl_link.visible = False
                self.color_picker_button.disable()
            openscad_str = self.code_area.value
            stl_path = stl_path = os.path.join(self.oscad.tmp_dir, self.stl_name)
            render_result = await self.oscad.openscad_str_to_file(
                openscad_str, stl_path
            )
            # show render result in log
            self.log_view.push(render_result.stderr)
            if render_result.returncode == 0:
                ui.notify("stl created ... loading into scene")
                self.stl_link.visible = True
                self.color_picker_button.enable()
                with self.scene:
                    self.stl_object = (
                        self.scene.stl(f"/stl/{self.stl_name}").move(x=0.0).scale(0.1)
                    )
                    self.stl_object.name = self.stl_name
                    self.stl_object.material(self.stl_color)
        except BaseException as ex:
            self.handle_exception(ex, self.do_trace)
        self.progress_view.visible = False

    def read_input(self, input_str: str):
        """Reads the given input and handles any exceptions.

        Args:
            input_str (str): The input string representing a URL or local file.
        """
        try:
            ui.notify(f"reading {input_str}")
            self.code = self.do_read_input(input_str)
            self.input_input.set_value(input_str)
            self.code_area.set_value(self.code)
            self.log_view.clear()
            self.error_msg = None
            self.stl_link.visible = False
        except BaseException as e:
            self.code = None
            self.handle_exception(e, self.do_trace)

    def save_file(self):
        """Saves the current code to the last input file, if it was a local path."""
        if self.is_local and self.input:
            with open(self.input, "w") as file:
                file.write(self.code)
            ui.notify(f"{self.input} saved")
        else:
            raise Exception("No local file to save to")

    async def open_file(self) -> None:
        """Opens a Local filer picker dialog and reads the selected input file."""
        if self.is_local:
            pick_list = await LocalFilePicker("~", multiple=False)
            if len(pick_list) > 0:
                input_file = pick_list[0]
                await self.read_and_optionally_render(input_file)

    pass

    def setup_pygments(self):
        """
        prepare pygments syntax highlighting by loading style
        """
        pygments_css_file = (
            Path(__file__).parent / "web" / "static" / "css" / "pygments.css"
        )
        pygments_css = pygments_css_file.read_text()
        ui.add_head_html(f"<style>{pygments_css}</style>")

    def code_changed(self, cargs):
        """
        react on changed code
        """
        self.code = cargs.value
        pass

    async def highlight_code(self, _cargs):
        """
        highlight the code and show the html
        """
        try:
            if self.code_area.visible:
                self.code_area.visible = False
                code_html = self.oscad.highlight_code(self.code)
                self.html_view.content = code_html
                self.html_view.visible = True
            else:
                self.html_view.visible = False
                self.code_area.visible = True
            self.toggle_icon(self.highlight_button)
        except BaseException as ex:
            self.handle_exception(ex, self.do_trace)

    async def pick_color(self, e: ColorPickEventArguments):
        """
        Asynchronously picks a color based on provided event arguments.

        This function changes the color of the 'color_picker_button' and the 'stl_object'
        according to the color specified in the event arguments.

        Args:
            e (ColorPickEventArguments): An object containing event-specific arguments.
                The 'color' attribute of this object specifies the color to be applied.

        Note:
            If 'stl_object' is None, the function will only change the color of 'color_picker_button'.
            Otherwise, it changes the color of both 'color_picker_button' and 'stl_object'.
        """
        self.color_picker_button.style(f"background-color:{e.color}!important")
        if self.stl_object:
            self.stl_color = e.color
            self.stl_object.material(f"{e.color}")
        pass

    async def toggle_axes(self):
        """
        toggle the axes of my scene
        """
        self.toggle_icon(self.axes_button)
        if self.axes_view is None:
            self.axes_view = AxesHelper(self.scene)
        else:
            self.axes_view.toggle_axes()
        pass

    async def toggle_grid(self, _ea):
        """
        toogle the grid of my scene
        """
        try:
            grid = self.scene._props["grid"]
            grid_str = "off" if grid else "on"
            grid_js = "false" if grid else "true"
            # try toggling grid
            ui.notify(f"setting grid to {grid_str}")
            grid = not grid
            # workaround according to https://github.com/zauberzeug/nicegui/discussions/1246
            js_cmd = f'scene_c{self.scene.id}.children.find(c => c.type === "GridHelper").visible = {grid_js}'
            await ui.run_javascript(js_cmd, respond=False)
            self.scene._props["grid"] = grid
            self.scene.update()
            # try toggling icon
            self.toggle_icon(self.grid_button)
        except BaseException as ex:
            self.handleExeption(ex)
        pass

    async def home(self):
        """Generates the home page with a 3D viewer and a code editor."""
        self.setup_pygments()
        self.setup_menu()
        with ui.column():
            with ui.splitter() as splitter:
                with splitter.before:
                    self.grid_button = self.tool_button(
                        "toggle grid",
                        handler=self.toggle_grid,
                        icon="grid_off",
                        toggle_icon="grid_on",
                    )
                    self.axes_button = self.tool_button(
                        "toggle axes",
                        icon="polyline",
                        toggle_icon="square",
                        handler=self.toggle_axes,
                    )
                    self.color_picker_button = ui.button(
                        icon="colorize", color=self.stl_color
                    )
                    with self.color_picker_button:
                        self.color_picker = ui.color_picker(on_pick=self.pick_color)
                    self.color_picker_button.disable()

                    with ui.scene(width=1024, height=768).classes("w-full") as scene:
                        self.scene = scene
                        scene.spot_light(distance=100, intensity=0.2).move(-10, 0, 10)
                    with splitter.after:
                        with ui.element("div").classes("w-full"):
                            extensions = {"scad": ".scad", "xml": ".xml"}
                            self.example_selector = FileSelector(
                                path=self.root_path,
                                handler=self.read_and_optionally_render,
                                extensions=extensions
                            )
                            self.input_input = ui.input(
                                value=self.input, on_change=self.input_changed
                            ).props("size=100")
                            self.highlight_button = self.tool_button(
                                tooltip="highlight",
                                icon="html",
                                toggle_icon="code",
                                handler=self.highlight_code,
                            )
                            if self.is_local:
                                self.tool_button(
                                    tooltip="save", icon="save", handler=self.save_file
                                )
                            self.tool_button(
                                tooltip="reload",
                                icon="refresh",
                                handler=self.reload_file,
                            )
                            if self.is_local:
                                self.tool_button(
                                    tooltip="open",
                                    icon="file_open",
                                    handler=self.open_file,
                                )
                            self.tool_button(
                                tooltip="render",
                                icon="play_circle",
                                handler=self.render,
                            )
                            self.stl_link = ui.link(
                                "stl result", f"/stl/{self.stl_name}", new_tab=True
                            )
                            self.stl_link.visible = False
                            self.progress_view = ui.spinner(
                                "dots", size="lg", color="blue"
                            )
                            self.progress_view.visible = False
                            self.code_area = (
                                ui.textarea(
                                    value=self.code, on_change=self.code_changed
                                )
                                .props("clearable")
                                .props("rows=25")
                            )
                            self.html_view = ui.html()
                            self.html_view.visible = False
                            self.log_view = ui.log(max_lines=20).classes("w-full h-40")
        await self.setup_footer()
        if self.args.input:
            await self.read_and_optionally_render(self.args.input)

    def configure_settings(self):
        """Generates the settings page with a link to the project's GitHub page."""
        sp_input = ui.textarea("scad prepend", value=self.oscad.scad_prepend).props(
            "cols=80"
        )
        sp_input.bind_value(self.oscad, "scad_prepend")

    def configure_run(self):
        root_path = (
            self.args.root_path if self.args.root_path else WebServer.examples_path()
        )
        self.root_path = os.path.abspath(root_path)
        self.allowed_urls = [
            "https://raw.githubusercontent.com/WolfgangFahl/nicescad/main/examples/",
            "https://raw.githubusercontent.com/openscad/openscad/master/examples/",
            self.examples_path(),
            self.root_path,
        ]



"""
Created on 2023-06-19

@author: wf
"""
import nicescad


class Version(object):
    """
    Version handling for nicescad
    """

    name = "nicescad"
    version = nicescad.__version__
    date = "2023-07-19"
    updated = "2023-12-17"
    description = "nicescad brings OpenScad to the browser (again)"

    authors = "Wolfgang Fahl"

    doc_url = "https://wiki.bitplan.com/index.php/nicescad"
    chat_url = "https://github.com/WolfgangFahl/nicescad/discussions"
    cm_url = "https://github.com/WolfgangFahl/nicescad"

    license = f"""Copyright 2023 contributors. All rights reserved.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied."""
    longDescription = f"""{name} version {version}
{description}

  Created by {authors} on {date} last updated {updated}"""



'''
Created on 2023-07-19

@author: wf
'''
import asyncio
import os
from tests.basetest import Basetest
from nicescad.openscad import OpenScad

class TestOpenScad(Basetest):
    """
    test openscad wrapper
    """
    
    def test_detect_openscad_exec(self):
        """
        test finding the openscad executable
        """
        oscad=OpenScad()
        self.assertTrue(oscad.openscad_exec is not None)
        
    def testRender(self):
        """
        test rendering a scad file
        """
        # openscad_exec="/Users/wf/bin/openscad"
        oscad=OpenScad()
        openscad_str="cube(5);"
        stl_path = os.path.join(oscad.tmp_dir, 'cube5.stl') 
        subprocess=asyncio.run(oscad.openscad_str_to_file(openscad_str,stl_path))
        debug=self.debug
        #debug=True
        if debug:
            print(subprocess)
        self.assertEqual(0,subprocess.returncode)
        self.assertEqual(stl_path,subprocess.stl_path)
        self.assertTrue(os.path.isfile(stl_path))
        pass
    
    def test_highlight_code(self):
        """
        Tests the 'highlight_code' function by checking if the output starts with 
        the standard beginning of an HTML string outputted by Pygments.
        
        This test case assumes the presence of a '<div' tag at the start of the 
        HTML string, which is standard for Pygments' HtmlFormatter. Adjust the 
        test case as necessary if you're using a different formatter or 
        configuration.
        
        Reference: https://github.com/WolfgangFahl/nicescad/issues/12
        
        """
        oscad=OpenScad()
        test_code = 'module test() { echo("Hello, world!"); }'
        highlighted_code = oscad. highlight_code(test_code)
        debug=self.debug
        if debug:
            print(highlighted_code)
        self.assertTrue(highlighted_code.startswith('<div'))


'''
Created on 2023-07-24

@author: wf
'''
from tests.basetest import Basetest
from nicescad.process import Subprocess

class TestSubprocess(Basetest):
    """
    test subprocess wrapper
    """
    
    def testPython(self):
        """
        test running a python command
        """
        #path="/opt/local/bin"
        cmd=[f"python","--version"]
        subprocess=Subprocess.run(cmd)
        debug=True
        if debug:
            print(subprocess)
        self.assertEqual(0,subprocess.returncode)
        


"""
A class to convert BlockSCAD XML files to SCAD files using the OpenAI language model API.

BlockSCAD Editor: https://www.blockscad3d.com/editor/
OpenSCAD: https://openscad.org/
NiceSCAD: http://nicescad.bitplan.com/
Relevant Issue: http://nicescad.bitplan.com/issue/23

Author: OpenAI ChatGPT
Date: July 25, 2023

The code was created by the OpenAI ChatGPT model in response to the following prompts:
- "Create a class named 'BlockscadConverter' that converts BlockSCAD XML files to SCAD files."
- "The xml_path is given in the constructor."
- "The class should have a method named `convert_to_scad` that outputs a path to the SCAD file."
- "Use the write method of a file object to output the SCAD file."
- "Include Google docstrings and type hints to the code."
- "Include comments that explain which prompts created the code. This helps other LLMs understand how to reproduce the same results."
- "Add links to the BlockSCAD editor (https://www.blockscad3d.com/editor/), OpenSCAD (https://openscad.org/), and the target platform (NiceSCAD, http://nicescad.bitplan.com/)."
- "Include the link to the relevant issue: 'support reading and converting blockscad files #23' on http://nicescad.bitplan.com/issue/23"
- "Add yourself (OpenAI ChatGPT) as the author and include the date"
- "keep the prompt list in the comments to be able to reproduce the results."
- "If the OpenAI API key is not available in the environment variables, look for it in a JSON file at `~/.openai/openai_api_key.json`."
- "If the OpenAI API key is not found, throw an exception."
"""

import json
import os
from pathlib import Path
from typing import Union

import openai


class BlockscadConverter:
    """
    Attributes
    ----------
    xml_path : str
        path to the input BlockSCAD XML file

    Methods
    -------
    convert_to_scad(scad_path: str) -> Union[str, None]
        Converts the BlockSCAD XML file to a SCAD file and returns the SCAD file path.
    """

    def __init__(self, xml_path: str):
        """
        Parameters
        ----------
        xml_path : str
            path to the input BlockSCAD XML file
        """
        self.xml_path = xml_path

    def convert_to_scad(self, scad_path: str) -> Union[str, None]:
        """
        Converts the BlockSCAD XML file to a SCAD file using the OpenAI language model API.

        Parameters
        ----------
        scad_path : str
            path to the output SCAD file

        Returns
        -------
        Union[str, None]
            path to the output SCAD file if conversion is successful, None otherwise
        """
        # Load the API key from the environment or a JSON file
        openai_api_key = os.getenv("OPENAI_API_KEY")
        json_file = Path.home() / ".openai" / "openai_api_key.json"

        if openai_api_key is None and json_file.is_file():
            with open(json_file, "r") as file:
                data = json.load(file)
                openai_api_key = data.get("OPENAI_API_KEY")

        if openai_api_key is None:
            raise ValueError(
                "No OpenAI API key found. Please set the 'OPENAI_API_KEY' environment variable or store it in `~/.openai/openai_api_key.json`."
            )

        openai.api_key = openai_api_key

        # Read the XML file
        with open(self.xml_path, "r") as file:
            xml_content = file.read()

        # Check if the XML content is a BlockSCAD XML
        if '<xml xmlns="https://blockscad3d.com' not in xml_content:
            msg = f"The file at {self.xml_path} is not a valid BlockSCAD XML file."
            raise Exception(msg)

        # Use the API to convert the XML to SCAD
        response = openai.Completion.create(
            engine="text-davinci-002",
            prompt=f"""Convert the following BlockSCAD XML to OpenSCAD and make sure to add a preamble comment (verbatim):
// OpenSCAD 
// converted from BlockSCAD XML by nicescad's blockscad converter
// according to 
// https://github.com/WolfgangFahl/nicescad/issues/23
// support reading and converting blockscad files #23
//
make sure to convert as direct as possible e.g. 
translate,rotate,cylinder,sphere,cube,color which are available in OpenScad should be           
used as such. 
Use all parameters e.g. prefer cube(10,10,10,center=true) to cube(10) when the parameters are available in
the original xml file.
<field name="CENTERDROPDOWN">false</field> e.g. leads to center=false
Add the original color command using a color name when applicable e.g. color("green");.
Try high reproduceability by not making any assumptions and keeping the structure intact. So do not add an initial translate. 

Do not add extra empty comments.
Avoid any empty lines - really - never add whitespace that harms getting always the same result.
Always use // line comments never /* 
Always indent with two spaces.
Make sure commands end with a semicolon.

Here is the BlockSCAD XML:\n{xml_content}
""",
            temperature=0.5,
            max_tokens=1000,
        )

        scad_content = response.choices[0].text.strip()

        # A very basic check to see if the SCAD content seems okay
        if not scad_content.startswith("// OpenSCAD"):
            msg = f"The conversion of {self.xml_path} failed - the // OpenSCAD comment is missing in:\n {scad_content}."
            raise Exception(msg)

        # Write the SCAD code to the output file
        with open(scad_path, "w") as scad_file:
            scad_file.write(scad_content)

        return scad_path



class AxesHelper:
    """
    A Python class for creating a 3D helper for visualizing the axes in a scene.
    This class is designed to be used with the `nicegui` library (https://pypi.org/project/nicegui/)
    for creating 3D scenes. It creates line objects for the x, y, and z axes,
    each with a distinct color, and provides methods for changing these colors and toggling their visibility.

    The original JavaScript code, which this Python code is based on, can be found at:
    https://raw.githubusercontent.com/mrdoob/three.js/master/src/helpers/AxesHelper.js
    The JavaScript code was refactored into this Python version by OpenAI's ChatGPT.

    For the refactoring, following prompts were given:
    1. Refactor a JavaScript class for handling ThreeJS scenes into a Python class using the `nicegui` library.
    2. The class should allow to hide/show axes.
    3. Usage of `nicegui` library's API for the colors.
    4. Include Google docstrings and type hints to the code.
    5. The scene should be a constructor parameter to be remembered.
    6. The Axes should be named x, y and z. The name attribute should be assigned after the line object creation.
    7. The `set_colors` method should be called in the constructor.
    8. Use `with self.scene as scene` when drawing the lines.
    9. The class should provide a method to toggle the visibility of axes.
    10. The default size of the axes to be drawn should be 10.0.
    11. Color information should be persisted when toggling the axes.
    12. Use a dictionary to store colors by axes names.

    According to nicegui's documentation (https://nicegui.io/documentation/scene), the 'scene.line()' function is used to
    create lines, here is an example:
        scene.line([-4, 0, 0], [-4, 2, 0]).material('#ff0000')
    To remove elements from the scene, one uses 'scene.remove()' function:
        scene.remove(element)
    where element is either the element instance or its ID.

    Date: 2023-07-24
    @author: Refactored into Python by OpenAI's ChatGPT

    Attributes:
        size (float): The size of the axes to be drawn.
        scene (ui.scene): The scene where the axes will be drawn.

    Usage:
        scene = ui.scene().classes('w-full h-64')
        axes_helper = AxesHelper(scene, size=10.0)
        axes_helper.set_colors(x_axis_color='#FF0000', y_axis_color='#00FF00', z_axis_color='#0000FF')  # set colors for x, y, and z axes
        axes_helper.toggle_axes()  # toggle visibility of the axes
    """

    def __init__(self, scene: "ui.scene", size: float = 10.0):
        """
        The constructor for AxesHelper class.

        Args:
            scene (ui.scene): The scene where the axes will be drawn.
            size (float): The size of the axes to be drawn.
        """
        self.scene = scene
        self.size = size
        self.vertices = [
            (0, 0, 0),
            (size, 0, 0),
            (0, 0, 0),
            (0, size, 0),
            (0, 0, 0),
            (0, 0, size),
        ]

        self.axis_names = ["x", "y", "z"]
        self.lines = []
        self.axes_visible = False
        self.color_by_name = {"x": "#FF0000", "y": "#00FF00", "z": "#0000FF"}
        self.toggle_axes()

    def set_colors(
        self,
        x_axis_color: str = "#FF0000",
        y_axis_color: str = "#00FF00",
        z_axis_color: str = "#0000FF",
    ):
        """
        A method to set colors of the axes.

        Args:
            x_axis_color (str): Color of the x-axis.
            y_axis_color (str): Color of the y-axis.
            z_axis_color (str): Color of the z-axis.
        """
        self.color_by_name = {"x": x_axis_color, "y": y_axis_color, "z": z_axis_color}
        for idx, line in enumerate(self.lines):
            line.material(self.color_by_name[self.axis_names[idx]])

    def toggle_axes(self):
        """
        A method to toggle the visibility of the axes.
        """
        if self.axes_visible:
            # Axes are currently visible, so remove them
            for line in self.lines:
                try:
                    self.scene.remove(line)
                except KeyError:
                    pass
            self.lines = []
            self.axes_visible = False
        else:
            # Axes are currently not visible, so add them
            with self.scene:
                for i in range(3):
                    line = self.scene.line(
                        self.vertices[2 * i], self.vertices[2 * i + 1]
                    )
                    line.name = self.axis_names[i]
                    line.material(self.color_by_name[line.name])
                    self.lines.append(line)
            self.axes_visible = True



"""
Created on 2023-07-19

@author: wf

This module contains the class OpenScad, a wrapper for OpenScad.
"""
import os
import platform
import tempfile
from typing import Awaitable

from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.lexer import RegexLexer
from pygments.token import (
    Comment,
    Keyword,
    Name,
    Number,
    Operator,
    Punctuation,
    String,
    Text,
)

from nicescad.process import Subprocess


class OpenSCADLexer(RegexLexer):
    """
    Lexer for OpenSCAD, a language for creating solid 3D CAD models.

    Attributes:
        name (str): The name of the lexer.
        aliases (list of str): A list of strings that can be used as aliases for the lexer.
        filenames (list of str): A list of strings that define filename patterns that match this lexer.
    """

    name = "OpenSCAD"
    aliases = ["openscad"]
    filenames = ["*.scad"]

    tokens = {
        "root": [
            (r"\s+", Text.Whitespace),
            (r"//.*?$", Comment.Single),
            (r"/\*.*?\*/", Comment.Multiline),
            (r"[a-z_][\w]*", Name.Variable),
            (r"\d+", Number.Integer),
            (r"\+\+|--", Operator),
            (r"[=+\-*/%&|^<>!]=?", Operator),
            (r"[\[\]{}();,.]", Punctuation),
            (r'"(\\\\|\\"|[^"])*"', String),
            (r"\b(module|if|else|for|let|echo)\b", Keyword),
            (r"\b(true|false|undef)\b", Keyword.Constant),
            (
                r"\b(cube|sphere|cylinder|polyhedron|square|circle|polygon|import|scale|resize|color|offset|minkowski|hull|render|surface|rotate|translate|mirror|multmatrix|projection|rotate_extrude|linear_extrude)\b",
                Name.Builtin,
            ),
        ],
    }


class OpenScad:
    """
    A wrapper for OpenScad (https://openscad.org/).
    """

    def __init__(self, scad_prepend: str = "", **kw) -> None:
        """
        Initializes the OpenScad object.
        """
        self.scad_prepend = scad_prepend
        self.openscad_exec = None
        self.openscad_tmp_dir = None
        if "OPENSCAD_EXEC" in os.environ:
            self.openscad_exec = os.environ["OPENSCAD_EXEC"]
        if "OPENSCAD_TMP_DIR" in os.environ:
            self.openscad_tmp_dir = os.environ["OPENSCAD_TMP_DIR"]
        if self.openscad_tmp_dir is not None:
            self.tmp_dir = self.openscad_tmp_dir
        else:
            self.tmp_dir = tempfile.mkdtemp()
        if "openscad_exec" in kw:
            self.openscad_exec = kw["openscad_exec"]
        if self.openscad_exec is None:
            self._try_detect_openscad_exec()
        if self.openscad_exec is None:
            raise Exception("openscad exec not found!")

    def highlight_code(self, code: str) -> str:
        """
        Highlights the provided OpenSCAD code and returns the highlighted code in HTML format.

        Args:
            code (str): The OpenSCAD code to highlight.

        Returns:
            str: The input OpenSCAD code, highlighted and formatted as an HTML string.
        """
        html = highlight(code, OpenSCADLexer(), HtmlFormatter())
        return html

    def _try_executable(self, executable_path: str) -> None:
        """
        Checks if the specified path is a file. If it is, sets it as the OpenScad executable.

        Args:
            executable_path (str): The path to the executable file.
        """
        if os.path.isfile(executable_path):
            self.openscad_exec = executable_path

    def _try_detect_openscad_exec(self) -> None:
        """
        Tries to find the OpenScad executable on the system.

        References:
            https://github.com/nickc92/ViewSCAD/blob/d4597ff6870316dfaafa4f9ecc8ef62773081c61/viewscad/renderer.py#L206C5-L222C1
        """
        platfm = platform.system()
        if platfm == "Linux":
            self._try_executable("/usr/bin/openscad")
            if self.openscad_exec is None:
                self._try_executable("/usr/local/bin/openscad")
        elif platfm == "Darwin":
            self._try_executable("/Applications/OpenSCAD.app/Contents/MacOS/OpenSCAD")
        elif platfm == "Windows":
            self._try_executable(
                os.path.join(
                    os.environ.get("Programfiles(x86)", "C:"), "OpenSCAD\\openscad.exe"
                )
            )
            self._try_executable(
                os.path.join(
                    os.environ.get("Programfiles", "C:"), "OpenSCAD\\openscad.exe"
                )
            )

    def write_to_tmp_file(self, openscad_str: str, do_prepend: bool = True):
        """
        Writes an OpenSCAD string to a temporary file.

        The `scad_prepend` string is prepended to the OpenSCAD code before writing,
        unless the OpenSCAD code contains the string '//!OpenSCAD', or `do_prepend`
        is set to `False`. In these cases, only the OpenSCAD code is written to the file.

        Args:
            openscad_str (str): The OpenSCAD code.
            do_prepend (bool, optional): If `True`, the `scad_prepend` string is
                                          prepended to the OpenSCAD code. Defaults to `True`.

        Returns:
            str: The path to the temporary file where the OpenSCAD code (and
                 possibly the `scad_prepend` string) was written.
        """
        scad_tmp_file = os.path.join(self.tmp_dir, "tmp.scad")
        with open(scad_tmp_file, "w") as of:
            if do_prepend and "//!OpenSCAD" not in openscad_str:
                of.write(self.scad_prepend)
            of.write(openscad_str)
        return scad_tmp_file

    async def render_to_file_async(
        self, openscad_str: str, stl_path: str
    ) -> Awaitable[Subprocess]:
        """
        Asynchronously renders an OpenSCAD string to a file.

        Args:
            openscad_str (str): The OpenSCAD code.
            stl_path(str): The path to the output file.

        Returns:
            Subprocess: the openscad execution result
        """
        scad_tmp_file = self.write_to_tmp_file(openscad_str)

        # now run openscad to generate stl:
        cmd = [self.openscad_exec, "-o", stl_path, scad_tmp_file]
        self.saved_umask = os.umask(0o077)
        result = await Subprocess.run_async(cmd)
        os.umask(self.saved_umask)

        self.cleanup_tmp_file(result, scad_tmp_file)
        result.stl_path = stl_path
        return result

    def cleanup_tmp_file(self, result, scad_tmp_file):
        """
        Cleanup temporary files after subprocess execution.

        Args:
            result (Subprocess): The result of the subprocess execution.
            scad_tmp_file (str): The path to the temporary file.
        """
        if result.returncode == 0:
            if os.path.isfile(scad_tmp_file):
                os.remove(scad_tmp_file)
        else:
            result.scad_tmp_file = scad_tmp_file

    async def openscad_str_to_file(
        self, openscad_str: str, stl_path: str
    ) -> Subprocess:
        """
        Renders the OpenSCAD code to a file.

        Args:
            openscad_str (str): The OpenSCAD code.
            stl_path(str): the path to the stl file

        Returns:
            Subprocess: The result of the subprocess run, encapsulated in a Subprocess object.
        """
        result = await self.render_to_file_async(openscad_str, stl_path)
        return result



'''
Created on 2023-07-23

@author: wf
'''
from tests.basetest import Basetest
from ngwidgets.file_selector import FileSelector
from nicescad.webserver import WebServer
import json

class TestFileSelector(Basetest):
    """
    test nicescad WebServer
    """
    
    def test_get_dir_tree(self):
        """
        test getting the directory tree structure for
        the examples directory
        """
        file_path=WebServer.examples_path()
        extensions = {"scad": ".scad", "xml": ".xml"}
        file_selector=FileSelector(file_path,extensions=extensions)
        debug=self.debug
        debug=True
        if debug:
            print(json.dumps(file_selector.tree_structure, indent=2))
        self.assertTrue("id" in file_selector.tree_structure)
        sample_id="1.3"
        node=file_selector.find_node_by_id(file_selector.tree_structure,sample_id)
        if debug:
            print(json.dumps(node,indent=2))
        self.assertIsNotNone(node)
        for key in ["id","label","value"]:
            self.assertTrue(key in node)


"""
Created on 2023-07-30

@author: wf
"""
import argparse

import uvicorn
from fastapi import FastAPI
from pydantic import BaseModel
from solid2 import *
from starlette.responses import HTMLResponse

import nicescad as nicescad


class SolidConverter:
    """
    Class for conversion of Python code to OpenSCAD code.
    """

    def __init__(self, python_code):
        self.python_code = python_code

    def convert_to_openscad(self):
        """
        Function to convert the input Python code into OpenSCAD code using SolidPython

        Returns:
        str -- OpenSCAD code
        """
        d = eval(self.python_code)
        openscad_code = scad_render(d)
        return openscad_code


class Item(BaseModel):
    python_code: str


class FastAPIServer:
    """
    Class for FastAPI server.
    """

    def __init__(self):
        self.app = FastAPI()
        self.app.post("/convert/")(self.convert)
        self.app.get("/version/")(self.version)
        self.app.get("/", response_class=HTMLResponse)(self.home)

    async def home(self):
        """
        Endpoint to return the homepage with links.

        Returns:
        str -- HTML content
        """
        return """
        <html>
            <head>
                <title>Nicescad solidpython converter service</title>
            </head>
            <body>
                <h1>Welcome to the nicescad solidpython to scad converter</h1>
                <ul>
                    <li><a href="/version/">Check the version of nicescad</a></li>
                    <li><a href="https://github.com/WolfgangFahl/nicescad/issues/28">nicescad GitHub issue</a></li>
                </ul>
            </body>
        </html>
        """

    async def version(self):
        """
        Endpoint to return the version of the nicescad package.

        Returns:
        dict -- the version
        """
        return {"version": nicescad.__version__}

    async def convert(self, item: Item):
        """
        Endpoint to convert Python code to OpenSCAD code.

        Arguments:
        item: Item -- input Python code

        Returns:
        dict -- the OpenSCAD code
        """
        converter = SolidConverter(item.python_code)
        openscad_code = converter.convert_to_openscad()
        return {"openscad_code": openscad_code}


def main():
    parser = argparse.ArgumentParser(
        description="Convert Python code to OpenSCAD code."
    )
    parser.add_argument(
        "--python_code", help="Python code to convert to OpenSCAD code."
    )
    parser.add_argument(
        "--file", help="File containing Python code to convert to OpenSCAD code."
    )
    parser.add_argument(
        "--serve", action="store_true", help="Start the FastAPI server."
    )
    parser.add_argument(
        "--host", default="0.0.0.0", help="Host address to bind the server to."
    )
    parser.add_argument(
        "--port", default=8000, type=int, help="Port number to bind the server to."
    )

    args = parser.parse_args()

    if args.serve:
        server = FastAPIServer()
        uvicorn.run(server.app, host="0.0.0.0", port=8000)
    else:
        if args.file:
            with open(args.file, "r") as f:
                python_code = f.read()
        elif args.python_code:
            python_code = args.python_code
        else:
            raise ValueError("Either python_code or file must be provided.")

        converter = SolidConverter(python_code)
        print(converter.convert_to_openscad())


if __name__ == "__main__":
    main()



import asyncio
import subprocess
from dataclasses import dataclass
from typing import Awaitable, List, Optional


@dataclass
class Subprocess:
    """A class representing a subprocess execution result."""

    stdout: str
    stderr: str
    cmd: List[str]
    returncode: int
    exception: Optional[BaseException] = None

    @staticmethod
    async def run_async(cmd: List[str]) -> Awaitable["Subprocess"]:
        """
        Asynchronously runs a command as a subprocess and returns the result as an instance of this class.

        Args:
            cmd (List[str]): The command to run.

        Returns:
            Subprocess: An instance of this class representing the result of the subprocess execution.
        """
        try:
            proc = await asyncio.create_subprocess_exec(
                *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )

            stdout, stderr = await proc.communicate()

            subprocess = Subprocess(
                stdout=stdout.decode(),
                stderr=stderr.decode(),
                cmd=cmd,
                returncode=proc.returncode,
            )
        except BaseException as ex:
            subprocess = Subprocess(
                stdout="", stderr=str(ex), cmd=cmd, returncode=-1, exception=ex
            )
        return subprocess

    @staticmethod
    def run(cmd: List[str]) -> "Subprocess":
        """
        Runs a command as a subprocess and returns the result as an instance of this class.

        Args:
            cmd (List[str]): The command to run.

        Returns:
            Subprocess: An instance of this class representing the result of the subprocess execution.
        """
        subprocess = asyncio.run(Subprocess.run_async(cmd))
        return subprocess



"""
Unit test for BlockscadConverter class to convert BlockSCAD XML files to SCAD files.

This test is based on the following prompts:
- "Conversion of a BlockSCAD XML file to a SCAD file."
- "Usage of nicescad.blockscad_converter.BlockscadConverter class for conversion."
- "The constructor needs an xml_path parameter."
- "The function to be tested is convert_to_scad with scad_path as the output file parameter."
- "Retrieval of the examples path using WebServer.examples_path() class method from nicescad.webserver."
- "Assumption of the 'blockscad' directory for SCAD input examples."
- "Definition of the unit test setup with paths for the BlockSCAD and SCAD directories."
- "Usage of test.basetest.Basetest as the base class for the test. Using def setUp(self, debug=False, profile=True):"
- "Definition of a test method to iterate over BlockSCAD XML files, convert them to SCAD files, and save in 'blockscad_converted' subdirectory."
- "Usage of a temporary directory for conversion during the test."
- "Comparison of output SCAD file content with the expected SCAD file content."
- "Inclusion of Google docstrings and type hints in the code."
- "Include links to the BlockSCAD editor (https://www.blockscad3d.com/editor/), OpenSCAD (https://openscad.org/), and the target platform (NiceSCAD, http://nicescad.bitplan.com/)."
- "Include the link to the relevant issue: 'support reading and converting blockscad files #23' on http://nicescad.bitplan.com/issue/23"
- "If self.debug is True, print the content of the compared files before doing the check."
- "Count the number of files and check that at least one file was tested."

Links for reference:
- BlockSCAD: https://www.blockscad3d.com/editor/
- OpenSCAD: https://openscad.org/
- Target platform (NiceSCAD): http://nicescad.bitplan.com/
- Issue: Support reading and converting blockcad files #23: http://nicescad.bitplan.com/issues/23

Author: OpenAI ChatGPT
Date: July 25, 2023
"""

import unittest
import tempfile
from tests.basetest import Basetest
from pathlib import Path
from nicescad.webserver import WebServer
from nicescad.blockscad_converter import BlockscadConverter

class TestBlockscadConverter(Basetest):
    def setUp(self, debug=True, profile=True):
        """
        Set up the test environment. Initializes paths for BlockSCAD XML files and expected SCAD files.

        Args:
            debug (bool): If True, prints the compared files before doing the check. Defaults to False.
            profile (bool): If True, the test is profiled. Defaults to True.
        """
        Basetest.setUp(self, debug, profile)
        self.examples_path = WebServer.examples_path()
        self.blockscad_dir = Path(self.examples_path) / 'blockscad'
        self.blockscad_converted_dir = Path(self.examples_path) / 'scad' / 'blockscad_converted'
    
    def compare_strings_ignore_whitespace(self,str1, str2):
        """
        Compares two strings for equality, ignoring any whitespace.
    
        Args:
            str1 (str): The first string to compare.
            str2 (str): The second string to compare.
    
        Returns:
            bool: True if the strings are equal when ignoring whitespace, False otherwise.
        """
        return ''.join(str1.split()) == ''.join(str2.split())
    
    def test_convert_to_scad(self):
        """
        Test the conversion of BlockSCAD XML files to SCAD files. Compares the content of the output SCAD files
        with the expected content.
        """
        if self.inPublicCI():
            return
        # only comment out when really testing since openai API usage
        # has a cost
        return
        blockscad_files = list(self.blockscad_dir.glob('*.xml'))
        self.assertGreater(len(blockscad_files), 0, "No BlockSCAD XML files found for testing.")
        failures=[]
        for xml_file in blockscad_files:
            converter = BlockscadConverter(str(xml_file))
            with tempfile.TemporaryDirectory() as temp_dir:
                scad_file = Path(temp_dir) / (xml_file.stem + '.scad')
                ok=False
                expected_scad_file = self.blockscad_converted_dir / (xml_file.stem + '.scad')
                try:
                    converter.convert_to_scad(str(scad_file))
                    with open(scad_file, 'r') as output_file, open(expected_scad_file, 'r') as expected_file:
                        output_content = output_file.read()
                        expected_content = expected_file.read()
                        ok = self.compare_strings_ignore_whitespace(output_content, expected_content)
                        if not ok:
                            print(f"Output content: \n{output_content}")
                            print(f"Expected content: \n{expected_content}")
                except Exception as ex:
                    print(str(ex))
                    pass            
                if not ok:
                    failures.append(expected_scad_file)
                            
        self.assertEqual(0,len(failures))                

if __name__ == '__main__':
    unittest.main()
